@@ -1,8 +1,8 @@
 # $File: //member/autrijus/ExtUtils-AutoInstall/AutoInstall.pm $ 
-# $Revision: #22 $ $Change: 2916 $ $DateTime: 2002/12/25 11:20:00 $
+# $Revision: #30 $ $Change: 4634 $ $DateTime: 2003/03/07 17:12:52 $
 
 package ExtUtils::AutoInstall;
-$ExtUtils::AutoInstall::VERSION = '0.44';
+$ExtUtils::AutoInstall::VERSION = '0.45';
 
 use strict;
 
@@ -15,23 +15,23 @@ ExtUtils::AutoInstall - Automatic install of dependencies via CPAN
 
 =head1 VERSION
 
-This document describes version 0.44 of B<ExtUtils::AutoInstall>,
-released December 25, 2002.
+This document describes version 0.45 of B<ExtUtils::AutoInstall>,
+released March 8, 2002.
 
 =head1 SYNOPSIS
 
 In F<Makefile.PL>:
 
-    # ExtUtils::AutoInstall Bootstrap Code, version 5.
-    BEGIN{my$p='ExtUtils::AutoInstall';my$v=0.40;eval"use $p $v;1
-    "or do{my$e=$ENV{PERL_EXTUTILS_AUTOINSTALL};(!defined($e)||$e
-    !~m/--(?:default|skip|testonly)/and-t STDIN or eval"use Ext".
-    "Utils::MakeMaker;WriteMakefile('PREREQ_PM'=>{'$p',$v});1"and
-    exit)and print"==> $p $v required. Install it from CPAN? [Y".
-    "/n] "and<STDIN>!~/^n/i and print"*** Installing $p\n"and do{
-    eval{require CPANPLUS;CPANPLUS::install $p};eval"use $p $v;1"
-    or eval{require CPAN;CPAN::install$p};eval"use $p $v;1"or die
-    "*** Please install $p $v manually from cpan.org first.\n"}}}
+    # ExtUtils::AutoInstall Bootstrap Code, version 6.
+    BEGIN{my$p='ExtUtils::AutoInstall';my$v=0.45;$p->VERSION||0>=$v
+    or+eval"use $p $v;1"or+do{my$e=$ENV{PERL_EXTUTILS_AUTOINSTALL};
+    (!defined($e)||$e!~m/--(?:default|skip|testonly)/and-t STDIN or
+    eval"use ExtUtils::MakeMaker;WriteMakefile(PREREQ_PM=>{'$p',$v}
+    );1"and exit)and print"==> $p $v required. Install it from CP".
+    "AN? [Y/n] "and<STDIN>!~/^n/i and print"*** Installing $p\n"and
+    do{eval{require CPANPLUS;CPANPLUS::install $p};eval("use $p $v;
+    1")||eval{require CPAN;CPAN::install$p};eval"use $p $v;1"or die
+    "*** Please manually install $p $v from cpan.org first...\n"}}}
 
     # optional pre-install handler; takes $module_name and $version
     # sub MY::preinstall  { return 1; }	# return false to skip install
@@ -46,6 +46,7 @@ In F<Makefile.PL>:
 	-config		=> {
 	    make_args	=> '--hello'	# option(s) for CPAN::Config
 	    force	=> 1,		# pseudo-option to force install
+	    do_once	=> 1,		# skip previously failed modules
 	},
 	-core		=> [		# core modules; may also be 'all'
 	    Package0	=> '',		# any version would do
@@ -98,6 +99,10 @@ B<ExtUtils::AutoInstall> lets module writers to specify a more
 sophisticated form of dependency information than the C<PREREQ_PM>
 option offered by B<ExtUtils::MakeMaker>.
 
+B<Module::Install> users should consult L<Module::Install::AutoInstall>
+for an alternative (and arguably more elegant) syntax to specify
+features, as demonstrated by this module's own F<Makefile.PL>.
+
 =head2 Prerequisites and Features
 
 Prerequisites are grouped into B<features>, and the user could choose
@@ -105,8 +110,8 @@ yes/no on each one's dependencies; the module writer may also supply a
 boolean value via C<-default> to specify the default choice.
 
 The B<Core Features> marked by the name C<-core> will double-check with
-the user, if the user chooses not to install the modules that belong to
-it.  This differs from the pre-0.26 'silent install' behaviour.
+the user, if the user chooses not to install the mandatory modules.
+This differs from the pre-0.26 'silent install' behaviour.
 
 Starting from version 0.27, if C<-core> is set to the string C<all>
 (case-insensitive), every feature will be considered mandatory.
@@ -163,7 +168,7 @@ B<Acme::KillarApp>, v1.21 or greater), then use something like below:
 
 Note the version test in the use clause; if you are so close to the
 cutting edge that B<Acme::KillerApp> 1.20 is the latest version on CPAN,
-this will prevent your module from go awry.
+this will prevent your module from going awry.
 
 =head2 User-Defined Hooks
 
@@ -195,9 +200,9 @@ module does not exist on CPAN at all.
 
 Starting from version 0.43, B<ExtUtils::AutoInstall> supports modules
 that require a C<MY::postamble> subroutine in their F<Makefile.PL>.
-The user-defined C<MY::postamble>, if present, is responsible to call
-C<ExtUtils::AutoInstall::postamble> and include the output in its
-return value.
+The user-defined C<MY::postamble>, if present, is responsible for
+calling C<ExtUtils::AutoInstall::postamble> and include the output in
+its return value.
 
 For example, the B<DBD::*> (database driver) modules for the Perl DBI
 are required to include the postamble generated by the function
@@ -337,7 +342,8 @@ sub import {
 
     while (my ($feature, $modules) = splice(@args, 0, 2)) {
 	my (@required, @tests, @skiptests);
-	my $default = 1;
+	my $default  = 1;
+	my $conflict = 0;
 
 	if ($feature =~ m/^-(\w+)$/) {
 	    my $option = lc($1);
@@ -367,6 +373,7 @@ sub import {
 		my $option = lc($1);
 
 		$default   = $arg    if ($option eq 'default');
+		$conflict  = $arg    if ($option eq 'conflict');
 		@tests     = @{$arg} if ($option eq 'tests');
 		@skiptests = @{$arg} if ($option eq 'skiptests');
 
@@ -375,6 +382,7 @@ sub import {
 
 	    printf("- %-${maxlen}s ...", $mod);
 
+	    # XXX: check for conflicts and uninstalls(!) them.
 	    if (defined(my $cur = _version_check(_load($mod), $arg ||= 0))) {
 		print "loaded. ($cur".($arg ? " >= $arg" : '').")\n";
 		push @Existing, $mod => $arg;
@@ -421,7 +429,7 @@ sub import {
 
     # import to main::
     no strict 'refs';
-    *{'main::WriteMakefile'} = \&Write;
+    *{'main::WriteMakefile'} = \&Write if caller(0) eq 'main';
 }
 
 # CPAN.pm is non-reentrant, so check if we're under it and have no CPANPLUS
@@ -453,31 +461,54 @@ sub install {
     my $i; # used below to strip leading '-' from config keys
     my @config = (map { s/^-// if ++$i; $_ } @{+shift});
 
-    my @modules;
-    my $installed = 0;
+    my (@modules, @installed);
     while (my ($pkg, $ver) = splice(@_, 0, 2)) {
 	# grep out those already installed
-	($installed++, next)
-	    if defined(_version_check(_load($pkg), $ver));
-	push @modules, $pkg, $ver;
+	if (defined(_version_check(_load($pkg), $ver))) {
+	    push @installed, $pkg;
+	}
+	else {
+	    push @modules, $pkg, $ver;
+	}
     }
 
-    return $installed unless @modules; # nothing to do
+    return @installed unless @modules; # nothing to do
 
     print "*** Installing dependencies...\n";
 
     return unless _connected_to('cpan.org');
 
+    my %args = @config;
+    my %failed;
+    local *FAILED;
+    if ($args{do_once} and open(FAILED, '.autoinstall.failed')) {
+	while (<FAILED>) { chomp; $failed{$_}++ }
+	close FAILED;
+	@modules = grep !$failed{$_}, @modules;
+    }
+
     if (_has_cpanplus()) {
-	$installed += _install_cpanplus(\@modules, \@config);
+	_install_cpanplus(\@modules, \@config);
     }
     else {
-	$installed += _install_cpan(\@modules, \@config);
+	_install_cpan(\@modules, \@config);
     }
 
     print "*** $class installation finished.\n";
 
-    return $installed;
+    # see if we have successfully installed them
+    while (my ($pkg, $ver) = splice(@modules, 0, 2)) {
+	if (defined(_version_check(_load($pkg), $ver))) {
+	    push @installed, $pkg;
+	}
+	elsif ($args{do_once} and open(FAILED, '>> .autoinstall.failed')) {
+	    print FAILED "$pkg\n";
+	}
+    }
+
+    close FAILED if $args{do_once};
+
+    return @installed;
 }
 
 sub _install_cpanplus {
@@ -504,6 +535,10 @@ sub _install_cpanplus {
     }
     $conf->set_conf(makeflags => $makeflags);
 
+    while (my ($key, $val) = splice(@config, 0, 2)) {
+	eval { $conf->set_conf($key, $val) };
+    }
+
     my $modtree = $cp->module_tree;
     while (my ($pkg, $ver) = splice(@modules, 0, 2)) {
 	print "*** Installing $pkg...\n";
@@ -520,7 +555,7 @@ sub _install_cpanplus {
 		delete $INC{$inc};
 	    }
 
-	    my $rv = $cp->install( modules => [ $obj->{module} ], @config);
+	    my $rv = $cp->install( modules => [ $obj->{module} ]);
 
 	    if ($rv and ($rv->{$obj->{module}} or $rv->{ok})) {
 		print "*** $pkg successfully installed.\n";
@@ -725,18 +760,7 @@ sub _version_check {
 # nothing; this usage is deprecated.
 sub main::PREREQ_PM { return {}; }
 
-# a wrapper to ExtUtils::MakeMaker::WriteMakefile
-sub Write {
-    require Carp;
-    Carp::croak "WriteMakefile: Need even number of args" if @_ % 2;
-
-    if ($CheckOnly) {
-	print << ".";
-*** Makefile not written in check-only mode.
-.
-	return;
-    }
-
+sub _make_args {
     my %args = @_;
 
     $args{PREREQ_PM} = { %{$args{PREREQ_PM} || {} }, @Existing, @Missing }
@@ -762,10 +786,27 @@ sub Write {
     ) if $Config;
 
     $PostambleActions = (
-	$missing ? "\$(PERL) $0 --config=$config --installdeps=$missing"
-		 : "\$(NOOP)"
+	$missing ? "\$(PERLRUN) $0 --config=$config --installdeps=$missing"
+		 : "\@\$(NOOP)"
     );
 
+    return %args;
+}
+
+# a wrapper to ExtUtils::MakeMaker::WriteMakefile
+sub Write {
+    require Carp;
+    Carp::croak "WriteMakefile: Need even number of args" if @_ % 2;
+
+    if ($CheckOnly) {
+	print << ".";
+*** Makefile not written in check-only mode.
+.
+	return;
+    }
+
+    my %args = _make_args(@_);
+
     no strict 'refs';
 
     $PostambleUsed = 0;
@@ -804,7 +845,7 @@ __END__
 =head1 SEE ALSO
 
 L<perlmodlib>, L<ExtUtils::MakeMaker>, L<Sort::Versions>, L<CPAN>,
-L<CPANPLUS>
+L<CPANPLUS>, L<Module::Install>
 
 =head1 ACKNOWLEDGEMENTS
 
@@ -821,7 +862,7 @@ Autrijus Tang E<lt>autrijus@autrijus.orgE<gt>
 
 =head1 COPYRIGHT
 
-Copyright 2001, 2002 by Autrijus Tang E<lt>autrijus@autrijus.orgE<gt>.
+Copyright 2001, 2002, 2003 by Autrijus Tang E<lt>autrijus@autrijus.orgE<gt>.
 
 This program is free software; you can redistribute it and/or
 modify it under the same terms as Perl itself.
